import gzip
import regex
import logging

import pandas as pd

from io import BytesIO


READCONTENTS = ['name', 'seq', 'spacer', 'quals']


def initialize_stats(nbcs):
    stats = {
        'valid': 0, 
        'filtered': 0,
        **{i: 0 for i in range(nbcs + 1)}
    }
    stats['poswise'] = [0] * nbcs
    return stats


def open_fastq(filepath):
    handle = (
        gzip.open(filepath, 'rb') 
        if filepath.endswith('gz') 
        else open(filepath, 'rb')
    )
    return handle


def get_read(fastq):
    read = {
        k: fastq.readline().rstrip()
        for k
        in READCONTENTS
    }
    
    if not all(read.values()):
        return {}

    # remove any unnecessary strings from spacer
    read['spacer'] = read['spacer'][:1]

    # this is necessary to ensure the aligner does not strip the barcodes later
    read['name'] = read['name'].split(maxsplit = 1)[0]
    return read


def read_fastqs(fastq1_path, fastq2_path):
    with (
        open_fastq(fastq1_path) as fastq1,
        open_fastq(fastq2_path) as fastq2
    ):
        read1 = get_read(fastq1)
        read2 = get_read(fastq2)
        while read1 and read2:
            yield read1, read2

            read1 = get_read(fastq1)
            read2 = get_read(fastq2)


def compress_read(read):
    string = b'\n'.join([read[k] for k in READCONTENTS]) + b'\n'
    return gzip.compress(string)


def increment_poswise_counter(read_bcs, stats):
    n_valid = 0
    for i, bc in enumerate(read_bcs):
        if not bc:
            continue
        
        n_valid += 1
        stats['poswise'][i] += 1

    return n_valid
    

def reads_to_byteblocks(reads):
    stats = {}
    bytestreams = dict(
        r1 = BytesIO(),
        filtered_r1 = BytesIO(),
        r2 = BytesIO(),
        filtered_r2 = BytesIO()
    )
    
    for read1, read2, bcs in reads:
        # initializing here to avoid passing number fo barcodes
        if not stats:
            nbcs = len(bcs)
            stats = initialize_stats(nbcs)

        n_valid_bcs = increment_poswise_counter(bcs, stats)
        bcs_string = b'|'.join(bcs)
        read1['name'] = read1['name'] + b'[' + bcs_string
        read2['name'] = read2['name'] + b'[' + bcs_string
        if not all(bcs):
            stats['filtered'] += 1
            stats[n_valid_bcs] += 1
            bytestreams['filtered_r1'].write(
                compress_read(read1)
            )
            bytestreams['filtered_r2'].write(
                compress_read(read2)
            )
            continue
        
        stats['valid'] += 1
        stats[nbcs] += 1
        bytestreams['r1'].write(
            compress_read(read1)
        )
        bytestreams['r2'].write(
            compress_read(read2)
        )

    return {k: stream.getvalue() for k, stream in bytestreams.items()}, stats


def write_byteblocks(byteblocks, outfilepaths):
    for k, block in byteblocks.items():
        outfile = outfilepaths[k]
        if not outfile:
            continue

        # need to use simple file here otherwise double compression
        with open(outfile, 'ab') as out:
            out.write(block)


def write_fastq(reads, outfilepaths):
    byteblocks, blockstats = reads_to_byteblocks(reads)
    write_byteblocks(
        byteblocks,
        outfilepaths
    )
    return blockstats


def read_barcodes(barcodes_path, allowed_mismatches):
    barcodes = pd.read_csv(
        barcodes_path,
        sep = '\t',
        header = None,
        names = ['category', 'bcname', 'bcseq']
    )
    bc_dict = {}
    max_bc_lengths = {}
    for cat, cat_barcodes in barcodes.groupby('category'):
        bc_lens = cat_barcodes.bcseq.str.len()
        max_bc_lengths[cat] = (bc_lens.min(), bc_lens.max())
        mismatches = allowed_mismatches[cat]
        bc_dict[cat] = {
            bytes(bc.bcseq, 'utf-8'): {
                'regex': regex.compile(
                    b'(' + bytes(bc.bcseq, 'utf-8') + 
                    b'){s<=' + 
                    bytes(str(mismatches), 'utf-8') + 
                    b'}'
                ), 
                'name': bytes(bc.bcname, 'utf-8'),
                'mismatch': mismatches
            }
            for _, bc
            in cat_barcodes.iterrows()
        }
    
    return bc_dict, max_bc_lengths


def write_overall_stats(stats_dict, statsfile):
    with open(statsfile, 'w') as file:
        for i, count in stats_dict.items():
            if i == 'poswise':
                continue

            file.write(f'{i}_barcodes\t{count}' + '\n')


def write_poswise_stats(poswise_counts, bc_cats, n_reads, statsfile):
    with open(statsfile, 'w') as file:
        header = '\t'.join(
            [
                f'{bc_cat}_{i}' 
                for bc_cat, i 
                in enumerate(bc_cats)
            ]
        )
        file.write(header + '\n')
        data = '\t'.join(
            [
                '{:.2f}'.format(c/n_reads * 100) 
                for c 
                in poswise_counts
            ]
        )
        file.write(data + '\n')


def sum_stats(stats, blockstats):
    if not blockstats:
        return
    
    for k, v in blockstats.items():
        if k == 'poswise':
            continue
        
        stats[k] += v

    for i, count in enumerate(blockstats['poswise']):
        stats['poswise'][i] += count


def initialize_stats_from_blockstats(blockstats):
    stats = {}
    for k, v in blockstats.items():
        if k == 'poswise':
            stats[k] = [0] * len(v)
            continue

        stats[k] = 0

    return stats


def write_parallel(
    outfilepaths,
    input_queue,
    lock,
    nextractors,
    bc_cats
):
    logging.info('starting writer process')
    stats = {}
    reads_processed = 0
    while True:
        reads = input_queue.get()
        if not reads:
            nextractors -= 1

        if not nextractors:
            break
        
        reads_processed += len(reads)
        byteblocks, blockstats = reads_to_byteblocks(reads)

        if not stats:
            stats = initialize_stats_from_blockstats(blockstats)

        sum_stats(stats, blockstats)

        # currently this is not necessary since we only use one writerthread
        # but we leave here for the future probably
        with lock:
            write_byteblocks(
                byteblocks,
                outfilepaths
            )
        
        if reads and not reads_processed % 1e5:
            logging.info(f'processed {reads_processed} reads')

    logging.info('all reads processed, shutting down writer')
    n_reads = stats['valid'] + stats['filtered']
    write_overall_stats(stats, outfilepaths['overall_stats'])
    write_poswise_stats(
        stats['poswise'], 
        bc_cats, 
        n_reads, 
        outfilepaths['poswise_stats']
    )


def initialize_output(outfilepaths):
    for path in outfilepaths.values():
        if path:
            open(path, 'w')


def parse_layout(layout_string, minmax_bc_len, allowed_mismatches):
    layout = []

    if not layout_string:
        return layout
    
    for bc_cat in layout_string.split('|'):
        min_len, max_len = minmax_bc_len[bc_cat]
        layout.append(
            [
                bc_cat,
                min_len,
                max_len,
                # this will return None in case of unknown categories
                # in case of e.g. SPACER this is okay since we anyway skip it
                allowed_mismatches.get(bc_cat) 
            ]
        )
    
    return layout


def parse_mismatches(mismatch_string):
    mismatch_dict = {}
    for cat_mismatch in mismatch_string.split(','):
        cat, mismatch = cat_mismatch.split(':')
        mismatch_dict[cat] = int(mismatch)

    return mismatch_dict


def add_spacer_info(minmax_bc_len, spacerlen):
    minmax_bc_len['SPACER'] = (spacerlen, spacerlen)
    


def add_pairs(subparser):
    parser = subparser.add_parser(
        'pairs',
        help = '''
        generate pairs files for each cluster size from alignments. Alignments have to be filtered such that
        they only contain valid alignments and no multimappers
        '''
    )
    parser.add_argument(
        '--bams',
        '-b',
        help = 'one or more BAM files containing aligned SPRITE-seq data',
        nargs = '+'
    )
    parser.add_argument(
        '--outprefix',
        '-o',
        help = 'prefix of the pairsfiles to write',
        required = True
    )
    parser.add_argument(
        '--clustersizelow',
        '-cl',
        help = 'minimum clustersize to consider',
        default = 2,
        type = int
    )
    parser.add_argument(
        '--clustersizehigh',
        '-ch',
        help = 'maximum clustersize to consider',
        default = 1000,
        type = int
    )
    parser.add_argument(
        '--separator',
        '-s',
        help = 'separator to use for extracting the barcode sequence from the readname',
        default = '['
    )
    parser.add_argument(
        '-ignoreprefix',
        help = 'prefixes of barcodes to ignore for building clusters in the form of a comma-separated list. e.g. <prefix1>,<prefix2>,...'
    )
    


def add_combine(subparser):
    parser = subparser.add_parser(
        'combine',
        help = '''
        merge multiple coolers generated from pairs of a given clustersize to a single one.
        This relies on the cooler names containing the clustersize.
        '''
    )
    parser.add_argument(
        '--input',
        '-i',
        help = 'path to directory containig clustersize coolers to merge. The name of the coolers must contain the clustersize _(?P<cs>[0-9]+)_',
        required = True
    )
    parser.add_argument(
        '--chunksize',
        help = 'number of rows of the pixel frame to fetch for merging',
        default = 1_000_000,
        type = int
    )
    parser.add_argument(
        '--floatcounts',
        help = 'if set stores count column as float, else stores them in a separate column count then contains rounded float counts',
        action = 'store_false',
        default = True
    )
    parser.add_argument(
        '--outfile',
        '-o',
        help = 'path to file where the merged Cooler should be written to',
        required = True
    )



import logging
from match import match_with_errors


def hash_match(seq, bc_dict, min_len, max_len):
    # if we only deal with one length we skip the loop
    if min_len == max_len:
        match = bc_dict.get(seq)
        return match['name'] if match else b'', max_len

    match_name = b''
    for bc_len in range(min_len, max_len + 1):
        match = bc_dict.get(seq[:bc_len])
        if match:
            match_name = match['name']
            break

    return match_name, bc_len


def extract_barcodes(read, bc_dicts, layout, laxity = 6):
    start = 0
    read_bcs = []

    if not layout:
        return read_bcs

    readseq = memoryview(read['seq'])
    # print(readseq)
    for bc_cat, min_bc_len, max_bc_len, allowed_mismatches in layout:
        # this is a shortcut to avoid matching the full SPACER cat
        if bc_cat.startswith('S'):
            # print(bc_cat, start, start + max_bc_len)
            # print(' '* start + readseq[start: start + max_bc_len])
            start += max_bc_len
            continue

        if not allowed_mismatches:
            bc_match, match_len = hash_match(
                readseq[start: start + max_bc_len],
                bc_dicts[bc_cat],
                min_bc_len,
                max_bc_len
            )
            # print(bc_cat, start, start + match_len)
            # print(' '* start + readseq[start: start + match_len])
            read_bcs.append(bc_match)
            start += match_len
            continue
        
        bc_match, match_pos = match_with_errors(
            readseq[start: start + max_bc_len + laxity].tobytes(),
            bc_dicts[bc_cat],
            laxity
        )
        # print(bc_cat, start, start + match_pos + max_bc_len)
        # print(' '* (start + match_pos)  + readseq[start + match_pos : start + match_pos + max_bc_len])
        read_bcs.append(bc_match)
        start += (match_pos + max_bc_len)

    return read_bcs


def extract_parallel(
    bc_dicts,
    layout1,
    layout2,
    laxity,
    input_queue, 
    output_queue
):
    logging.info('start extractor process')
    while True:
        reads = input_queue.get()
        if not reads:
            break
        
        for readpair in reads:
            read1, read2 = readpair
            readpair.append(
                extract_barcodes(read1, bc_dicts, layout1, laxity) +
                extract_barcodes(read2, bc_dicts, layout2, laxity)
            )

        output_queue.put(reads)

    # termination signal
    logging.info('received empty readlist, shutting down extractor')
    output_queue.put([])



from cooler import fileops, create

import h5py
import logging

import pandas as pd


def attrs_to_dict(h5file):
    attrs = {
        key: h5file.attrs[key]
        for key
        in h5file.attrs.keys()
    }
    return attrs


def copy_attrs(source, dest):
    with (
        h5py.File(source, 'r+') as infile,
        h5py.File(dest, 'r+') as ofile
    ):
        attrs = attrs_to_dict(infile)
        ofile.attrs.update(attrs)


def to_right_dtype(x):
    try:
        return int(x)
    
    except ValueError:
        return x


def read_bed_by_chrom(bedfile):
    # this relies on the bed being sorted
    linebuffer = []
    with open(bedfile, 'r') as bed:
        current_chrom = None
        for line in bed:
            line = [
                to_right_dtype(field) 
                for field 
                in line.rstrip().split('\t')
            ]

            if not current_chrom:
                current_chrom = line[0]
            
            if current_chrom != line[0]:
                logging.info(f'processed {current_chrom}')
                chrom_data = pd.DataFrame(
                    linebuffer,
                    columns = ['chrom', 'start', 'end', 'name']
                )
                current_chrom = line[0]
                linebuffer = [line]
                yield chrom_data
                continue

            linebuffer.append(line)
        
        logging.info(f'processed {current_chrom}')
        chrom_data = pd.DataFrame(
            linebuffer,
            columns = ['chrom', 'start', 'end', 'name']
        )
        yield chrom_data


def get_h5_group(h5, key_sequence):
    grp = h5
    for key in key_sequence:
        grp = grp[key]
    
    return grp


def write_annotation(grp, name, data, h5opts):
    if name in grp:
        del grp[name]
    
    grp.create_dataset(name, data=data, **h5opts)


def copy_and_annotate_cooler(source, dest, annotations, h5opts = None, mcoolfile = True):
    fileops.cp(source, dest)
    keys = ['/']
    ofile = dest

    if mcoolfile:
        ofile, keystring = dest.split('::')
        keys.extend(keystring[1:].split('/'))
        
    with h5py.File(ofile, 'r+') as h5:
        rootgrp = get_h5_group(h5, keys)
        grp = rootgrp['bins']
        h5opts = create._create._set_h5opts(h5opts)
        for col in annotations.columns:
            write_annotation(grp, col, annotations[col].values, h5opts)
    


import logging
import gc

from .ioutils import (
    get_resolutons,
    check_weight,
    store_weights,
    copy_attrs
)
from .core import (
    balance_ic,
    balance_kr
)
from cooler import fileops

def main(args):
    for coolpath in fileops.list_coolers(args.mcool):
        cooleruri = args.mcool + '::' + coolpath
        outuri = args.output + '::' + coolpath
        fileops.cp(cooleruri, outuri)

        for weight_name, per_chrom, balancefunc in zip(
            ['KR', 'perChromKR', 'ICE', 'perChromIC'],
            [False, True, False, True],
            [balance_kr, balance_kr, balance_ic, balance_ic]
        ):
            if not check_weight(cooleruri, weight_name) or args.overwrite:
                logging.info(
                    'computing {} for {}::{}'.format(
                        weight_name, 
                        args.mcool, 
                        coolpath
                    )
                )
                weights, stats = balancefunc(
                    outuri, 
                    per_chrom,
                    nproc = args.processors,
                    maxiter = args.maxiter
                )
                store_weights(
                    outuri, 
                    weights, 
                    weight_name,
                    stats,
                    overwrite = args.overwrite
                )
                del weights

            else:
                logging.info(
                    '{} weights for {}::{} already exist. Skipping!'.format(
                        weight_name, 
                        args.mcool, 
                        coolpath
                    )
                )

        gc.collect()

    copy_attrs(args.mcool, args.output)



import logging
import cooler

import numpy as np
import multiprocessing as mp

from krbalancing import *
from .ioutils import cooler_to_csr


def remove_nan_bin_weights(csrmatrix, weights):
    '''
    removes weights from the balancing weight vector that belong to
    bins with 0 coverage (i.e. NaN bins)

    :param csrmatrix:   cooler as scipy.sparse.csr_matrix
    :param weights:     numpy.array of weights

    :return:            processed weight vector
    '''
    nnz_rows = csrmatrix.getnnz(1)
    nnz_cols = csrmatrix.getnnz(0)
    weights[(nnz_rows == 0) & (nnz_cols == 0)] = np.nan
    return weights


def balance_kr(cooleruri, per_chrom = False, **kwargs):
    '''
    applies KR matrix balancing to the given cooleruri
    code taken from HiCExplorer's hicCorrectMatrix see also
    https://github.com/deeptools/HiCExplorer/blob/master/hicexplorer/hicCorrectMatrix.py

    :param cooleruri:   uri to a given cooler

    :return:            KR balancing weights
    '''
    csrmatrix = cooler_to_csr(cooleruri)

    def apply_kr(m):
        kr = kr_balancing(
            m.shape[0],
            m.shape[1],
            m.count_nonzero(),
            m.indptr.astype(np.int64, copy=False),
            m.indices.astype(np.int64, copy=False),
            m.data.astype(np.float64, copy=False)
        )
        kr.computeKR()

        # set it to False since the vector is already normalised
        # with the previous True
        # correction_factors = np.true_divide(1, kr.get_normalisation_vector(False).todense())
        w = kr.get_normalisation_vector(False).todense()

        # flatten weights to comply with cooler format specification
        return np.array(w).flatten()

    if not per_chrom:
        weights = apply_kr(csrmatrix)

    else:
        coolermatrix = cooler.Cooler(cooleruri)
        weights = np.array([])
        for chrom in coolermatrix.chromnames:
            chromextent = coolermatrix.extent(chrom)
            chromcsr = csrmatrix[chromextent[0]: chromextent[1], chromextent[0]: chromextent[1]]
            chromweights = apply_kr(chromcsr)
            weights = np.concatenate([weights, np.array(chromweights).flatten()])

    return remove_nan_bin_weights(csrmatrix, weights), {'divisive_weights': False}


def balance_ic(cooleruri, per_chrom, nproc, maxiter):
    '''
    applies IC matrix balancing to a given cooleruri
    code taken from cooler's cooler balance see also
    https://github.com/mirnylab/cooler/blob/master/cooler/cli/balance.py
    :param cooleruri:   uri to a given cooler
    :param nproc:       number of processors to use for balancing

    :return:            IC balancing weights
    '''
    clr = cooler.Cooler(cooleruri)
    try:
        if nproc > 1:
            pool = mp.Pool(nproc)
            map_ = pool.imap_unordered
        else:
            map_ = map

        bias, stats = cooler.balance_cooler(
            clr,
            chunksize=int(10e6),
            cis_only=per_chrom,
            trans_only=False,
            tol=1e-5,
            min_nnz=10,
            min_count=0,
            blacklist=None,
            mad_max=5,
            max_iters=maxiter,
            ignore_diags=2,
            rescale_marginals=True,
            use_lock=False,
            map=map_
        )

    finally:
        if nproc > 1:
            pool.close()

    # per chrom returns a bool for each balanced chrom
    converged = stats['converged'] if not per_chrom else stats['converged'].all()
    if not converged:
        logging.error('Iteration limit reached without convergence')
        logging.error('Storing final result. Check log to assess convergence.')

    return bias, stats



import glob
import re
import os
import resource
import logging

import pandas as pd

from cooler import Cooler


cs_regex = re.compile('_(?P<cs>[0-9]+)_')


def check_file_limit(n_files):
    softlimit, hardlimit = resource.getrlimit(
        resource.RLIMIT_NOFILE
    )

    required = n_files + 10
    if softlimit < n_files:
        logging.info(
            'open file limit is too small for current merger. ' +
            f'Is soft = {softlimit}, hard = {hardlimit}. ' +
            f'Needs to be at least {n_files}. Setting to {required} temporarily'
        )
        softlimit = required
        hardlimit = required

        resource.setrlimit(
            resource.RLIMIT_NOFILE,
            (softlimit, hardlimit)
        )


def clustersize_from_filename(filename):
    m = cs_regex.search(filename)
    return int(m.group('cs'))


def read_coolers(directory):
    coolers = {}
    for coolfile in glob.glob(directory + '/*'):
        clustersize = clustersize_from_filename(
            os.path.basename(coolfile)
        )
        coolers[clustersize] = Cooler(coolfile)
    
    return coolers



import logging

import pandas as pd
import numpy as np

from functools import partial
from cooler._reduce import CoolerMerger, merge_breakpoints


def load_and_correct(cool, size, lo, hi):
    selector = cool.pixels()
    pixels = selector[lo: hi]
    pixels['count'] *= (2/size)
    return pixels


class SpriteCoolerMerger(CoolerMerger):
    def __init__(self, coolers, mergebuf):
        super().__init__(list(coolers.values()), mergebuf)
        self.coolers = coolers

    def __iter__(self):
        # Load bin1_offset indexes lazily.
        indexes = [c.open("r")["indexes/bin1_offset"] for c in self.coolers.values()]

        # Calculate the common partition of bin1 offsets that define the epochs
        # of merging data.
        bin1_partition, cum_nrecords = merge_breakpoints(indexes, self.mergebuf)
        nrecords_per_epoch = np.diff(cum_nrecords)
        nnzs = [len(c.pixels()) for c in self.coolers.values()]

        starts = [0] * len(self.coolers)
        for i, bin1_id in enumerate(bin1_partition[1:], 2):
            stops = [index[bin1_id] for index in indexes]
            # extract, concat
            combined = pd.concat(
                [
                    load_and_correct(c, size, start, stop)
                    for (size, c), start, stop in zip(self.coolers.items(), starts, stops)
                    if (stop - start) > 0
                ],
                axis=0,
                ignore_index=True,
            )

            # sort and aggregate
            df = combined \
                .groupby(["bin1_id", "bin2_id"], sort=True) \
                .sum() \
                .reset_index()

            percent_merged = np.floor(i / len(bin1_partition) * 100)
            if not percent_merged % 10:
                logging.info(f'merged approximately {percent_merged} %')

            yield {k: v.values for k, v in df.items()}
            starts = stops



def add_balance(subparser):
    parser = subparser.add_parser(
        'balance',
        help = '''
        balance multi resolution coolers with KR and ICE algorithms
        '''
    )
    parser.add_argument(
        '-m', '--mcool',
        required = True,
        help = 'MultiCooler file to balance'
    )
    parser.add_argument(
        '-p', '--processors',
        default = 1,
        type = int,
        help = 'number of processors to use for IC balancing'
    )
    parser.add_argument(
        '--output',
        '-o',
        required = True,
        help = 'name of the output mcool file. input is copied there and appended with the weights'
    )
    parser.add_argument(
        '--overwrite',
        help = 'if set overwrites existing weight columns',
        default = False,
        action = 'store_true'
    )
    parser.add_argument(
        '--maxiter',
        default = 1000,
        help = 'maximum iterations for ICE algorithm'
    )



from .ioutils import (
    read_barcodes,
    parse_layout,
    parse_mismatches,
    add_spacer_info
)
from .processors import (
    process_parallel, 
    process_sequential
)
import logging


def generate_fileprefix(filename):
    prefix_tokens = filename.split('.')
    slice_end = -2 if filename.endswith('gz') else -1
    return '.'.join(prefix_tokens[:slice_end])
    

def main(args):
    logging.info('reading barcodes, layouts and setting up output files')
    allowed_mismatches = parse_mismatches(args.mismatches)
    bc_dicts, minmax_bc_len = read_barcodes(args.barcodes, allowed_mismatches)
    # add spacer info so we don't have to explicitly care about it
    add_spacer_info(minmax_bc_len, args.spacerlen)

    layout_r1 = parse_layout(args.layout1, minmax_bc_len, allowed_mismatches)
    layout_r2 = parse_layout(args.layout2, minmax_bc_len, allowed_mismatches)

    fileprefix = generate_fileprefix(args.output)
    outfilepaths = dict(
        r1 = args.output,
        filtered_r1 = fileprefix + '.filtered.fq.gz' if args.writefiltered else None,
        r2 = fileprefix + '.r2.fq.gz' if args.writer2 else None,
        filtered_r2 = fileprefix + '.filtered_r2.fq.gz' if args.writer2 and args.writefiltered else None,
        overall_stats = fileprefix + '.overall.stats.tsv',
        poswise_stats = fileprefix + '.poswise.stats.tsv'
    )

    if args.processes > 1:
        logging.info('processing in parallel')
        process_parallel(
            args.read1,
            args.read2,
            outfilepaths,
            bc_dicts,
            layout_r1,
            layout_r2,
            args.laxity,
            args.processes
        )

    else:
        logging.info('processing sequentially')
        process_sequential(
            args.read1,
            args.read2,
            outfilepaths,
            bc_dicts,
            layout_r1,
            layout_r2,
            args.laxity
        )
        


cdef bint match(
    const unsigned char* string, 
    const unsigned char* pattern, 
    const int allowed_mismatches
):
    cdef int n_mismatch = 0
    cdef int i = 0
    cdef int pattern_len = len(pattern)
    cdef int string_len = len(string)

    if string_len < pattern_len:
        return False

    for i in range(pattern_len):
        if not string[i] == pattern[i]:
            n_mismatch += 1
        
        if n_mismatch > allowed_mismatches:
            return False

    return True


cdef find_match(
    const unsigned char* seq, 
    dict bc_dict
):
    cdef char* match_name = b''
    cdef const unsigned char* bc
    cdef dict bcinfo
    cdef bint m
    for bc, bcinfo in bc_dict.items():
        m = match(seq, bc, bcinfo['mismatch'])
        if m:
            match_name = bcinfo['name']
            break
        
    return match_name


def match_with_errors(
    const unsigned char* seq, 
    dict bc_dict, 
    const int laxity
):
    match = None
    for match_start in range(laxity):
        match = find_match(seq[match_start:], bc_dict)
        if match:
            break

    return match, match_start



import logging

import numpy as np

from .ioutils import read_coolers, check_file_limit
from .core import SpriteCoolerMerger
from cooler.create import create


def main(args):
    logging.info('reading coolers')
    coolers = read_coolers(args.input)
    check_file_limit(len(coolers))
    logging.info('merging coolers')
    if not args.floatcounts:
        columns = ['count', 'float_count']
        dtypes = None
        logging.info('storing float counts in extra column')
    
    else:
        columns = None
        dtypes = {'count': float}
        logging.info('storing count as float')

    key = np.random.choice(list(coolers.keys()))
    bins = coolers[key].bins()[:]
    assembly = coolers[key].info.get("genome-assembly", None)
    iterator = SpriteCoolerMerger(
        coolers, 
        mergebuf = args.chunksize
    )

    create(
        args.outfile,
        bins,
        iterator,
        columns=columns,
        dtypes=dtypes,
        assembly=assembly
    )



from .extract import (
    extract_parallel,
    extract_barcodes
)

from .ioutils import (
    read_fastqs, 
    write_fastq,
    write_parallel,
    write_overall_stats,
    write_poswise_stats,
    initialize_stats,
    initialize_output,
    sum_stats
)

import multiprocessing as mp

import logging

BUFFERSIZE = 10000
                

def process_sequential(
    r1file, 
    r2file, 
    outfilepaths,
    bc_dicts,
    layout_r1,
    layout_r2,
    laxity
):
    stats = {}
    read_buffer = []
    reads_processed = 0
    for read1, read2 in read_fastqs(r1file, r2file):
        bcs = (
            extract_barcodes(read1, bc_dicts, layout_r1, laxity) +
            extract_barcodes(read2, bc_dicts, layout_r2, laxity)
        )

        if not stats:
            stats = initialize_stats(len(bcs))

        bcs_string = b'|'.join(bcs)
        read1['name'] = read1['name'] + bcs_string
        read2['name'] = read2['name'] + bcs_string
        read_buffer.append(
            [read1, read2, bcs]
        )

        if len(read_buffer) == BUFFERSIZE:
            blockstats = write_fastq(
                read_buffer,
                outfilepaths
            )
            sum_stats(stats, blockstats)

            read_buffer = []

        reads_processed += 1
        if not reads_processed % 1e5:
            logging.info(f'processed {reads_processed} reads')
    
    # write last reads
    blockstats = write_fastq(
        read_buffer,
        outfilepaths
    )
    sum_stats(stats, blockstats)

    bc_cats = [c[0] for c in layout_r1 + layout_r2 if c[0] != 'SPACER']
    n_reads = stats['valid'] + stats['filtered']
    write_overall_stats(stats, outfilepaths['overall_stats'])
    write_poswise_stats(
        stats['poswise'], 
        bc_cats, 
        n_reads, 
        outfilepaths['poswise_stats']
    )


def process_parallel(
    r1file, 
    r2file, 
    outfilepaths,
    bc_dicts,
    layout_r1,
    layout_r2,
    laxity,
    nprocesses
):
    # set queue size to avoid loading the full file in memory
    # queue blocks when full until space is freed
    extract_queue = mp.Queue(maxsize = 100)
    write_queue = mp.Queue(maxsize = 100)
    lock = mp.Lock()
    initialize_output(outfilepaths)

    extractor_args = (
        bc_dicts,
        layout_r1,
        layout_r2,
        laxity,
        extract_queue,
        write_queue
    )
    extractors = [
        mp.Process(
            target = extract_parallel,
            args = extractor_args
        )
        for _ in range(max(nprocesses - 2, 1))
    ]
    bc_cats = [c[0] for c in layout_r1 + layout_r2 if c[0] != 'SPACER']
    writer = mp.Process(
        target = write_parallel,
        args = (
            outfilepaths,
            write_queue,
            lock,
            len(extractors),
            bc_cats
        )
    )

    parallel_processes = [*extractors, writer]
    for process in parallel_processes:
        process.start()

    read_buffer = []
    for read1, read2 in read_fastqs(r1file, r2file):
        read_buffer.append([read1, read2])

        if len(read_buffer) == BUFFERSIZE:
            extract_queue.put(read_buffer)
            read_buffer = []

    # send last reads
    extract_queue.put(read_buffer)

    for _ in extractors:
        extract_queue.put([])

    logging.info('waiting for processes to finish')
    for process in parallel_processes:
        process.join()

    logging.info('all done!')



import logging

from .ioutils import (
    read_bam,
    clusters_to_pairs,
    write_stats
)
from collections import defaultdict


def main(args):
    clusters = defaultdict(set)
    logging.info('reading alignments and constructing clusters')
    alignments_processed, duplicate_alignments = 0, 0
    for bcs, pos in read_bam(args.bams, args.separator, args.ignoreprefix):
        if pos in clusters[bcs]:
            duplicate_alignments += 1
        
        clusters[bcs].add(pos)
        alignments_processed += 1

        if not alignments_processed % 1e5:
            logging.info(f'processed {alignments_processed} alignments')

    logging.info('finished cluster construction. write duplication stats')
    alignment_stats = {
        'unique': alignments_processed - duplicate_alignments,
        'duplicated': duplicate_alignments
    }
    write_stats(
        alignment_stats, 
        args.outprefix + '.duplicatestats.tsv'
    )
    logging.info('now writing pairs for all found sizes')
    size_stats = clusters_to_pairs(
        clusters, 
        args.outprefix,
        args.clustersizelow,
        args.clustersizehigh
    )
    logging.info('finished writing pairs. writing stats')
    write_stats(
        size_stats, 
        args.outprefix + '.sizestats.tsv'
    )



from pybedtools import BedTool
from .ioutils import read_bed_by_chrom

import pandas as pd

import os
import logging


def annotate_bins(cool, clusterbedfile):
    cool_chroms = set(cool.chromnames)
    annotated_bins = []
    for chrom_bed in read_bed_by_chrom(clusterbedfile):
        chrom = chrom_bed.iloc[0, 0]
        if not chrom in cool_chroms:
            continue
        
        lo, hi = cool.extent(chrom)
        chrom_bins = cool.bins()[lo:hi].loc[:, ['chrom', 'start', 'end']]
        chrom_bins['name'] = 'bin'
        a = BedTool.from_dataframe(chrom_bed)
        b = BedTool.from_dataframe(chrom_bins)
        intersect = b.intersect(a, wao = True, sorted = True).to_dataframe()
        grouped = intersect[['chrom', 'start', 'end', 'thickEnd']].groupby(['chrom', 'start', 'end'])
        annotated_bins.append(
            grouped.agg({'thickEnd': ','.join}).reset_index()
        )
    
    annotation = pd.concat(annotated_bins, ignore_index = True)
    colname = os.path.basename(clusterbedfile)
    annotation.rename(
        columns = {'thickEnd': colname},
        inplace = True
    )
    return annotation



def add_annotate(subparser):
    parser = subparser.add_parser(
        'annotate',
        help = '''
        annotate bins of one or multiple coolers with cluster information.
        i.e. if a read of a given cluster overlaps a given bin. annotated data is written to new file using old filename as prefix
        '''
    )    
    parser.add_argument(
        '--input',
        '-i',
        help = 'cooler (possibly mcool) to annotate with respective clusterinfo',
        required = True
    )
    parser.add_argument(
        '--bed',
        '-b',
        nargs = '+',
        help = '''
        one or more BEDfiles containing all valid SPRITE reads annotated with their cluster membership. 
        must be sorted by chrom, start. annotation columns will be named after bed basenames
        ''',
        required = True
    )
    parser.add_argument(
        '--outprefix',
        '-o',
        required = True,
        help = 'file prefix to use for writing resulting annotation'
    )
    


import re

import pysam as ps
import itertools as it

from collections import defaultdict, namedtuple

Position = namedtuple('Position', ['chrom', 'start', 'end', 'strand'])
def read_to_pos(read, sep, ignore_regex):
    bcs = read.query_name.split(sep)[-1]
    if not ignore_regex:
        bc_string = bcs

    else:
        # make use of list comp because faster
        bc_string = '|'.join(
            bc for bc in bcs.split('|') 
            if not any(r.match(bc) for r in ignore_regex)
        )

    pos = Position(
        chrom = read.reference_name, 
        start = read.reference_start,
        end = read.reference_end,
        strand = '+' if not read.is_reverse else '-'
    )
    return bc_string, pos


def parse_ignore_prefixes(prefixstring):
    return [re.compile(f'^{prefix}') for prefix in prefixstring.split(',')]


def read_bam(bamfiles, bcsep, ignore_prefixes):
    ignore_regex = parse_ignore_prefixes(ignore_prefixes) if ignore_prefixes else []
    for bamfile in bamfiles:
        with ps.AlignmentFile(bamfile, 'rb') as bam:
            # until_eof relies on filtering unmapped reads beforehand
            for read in bam.fetch(until_eof = True):
                yield read_to_pos(read, bcsep, ignore_regex)


def clusters_by_size(clusters):
    sized_clusters = defaultdict(list)
    for k, cluster in clusters.items():
        sized_clusters[len(cluster)].append([k, cluster])

    return sized_clusters


def write_pairs(clusters_of_size, filename, size):
    with open(filename, 'w') as file:
        for i, (k, cluster) in enumerate(clusters_of_size):
            for positions in it.combinations(cluster, 2):
                # ensures upper triangle
                pos1, pos2 = sorted(positions)
                #columns: readID chr1 pos1 chr2 pos2 strand1 strand2
                pair = [
                    # +1 because pairs is 1-based
                    pos1.chrom, pos1.start + 1, 
                    pos2.chrom, pos2.start + 1, 
                    pos1.strand, pos2.strand
                ]
                file.write(
                    '\t'.join([f'c_{size}_{i}', *[str(p) for p in pair]]) + '\n'
                )


def clusters_to_bed(size, clusters_of_size, file):
    for i, (k, cluster) in enumerate(clusters_of_size):
        for pos in cluster:
            file.write(
                '\t'.join(
                    [str(pos.chrom), str(pos.start), str(pos.end), f'c_{size}_{i}']
                ) + '\n'
            )


def write_bed(sized_clusters, filename, min_c_size, max_c_size):
    with open(filename, 'w') as file:
        for size, clusters_of_size in sized_clusters.items():
            if size < min_c_size or size > max_c_size:
                continue

            clusters_to_bed(
                size,
                clusters_of_size,
                file
            )


def clusters_to_pairs(clusters, fileprefix, min_c_size, max_c_size):
    sized_clusters = clusters_by_size(clusters)
    stats = {}
    for size, clusters_of_size in sized_clusters.items():
        stats[size] = len(clusters_of_size)
        if size < min_c_size or size > max_c_size:
            continue
        
        write_pairs(
            clusters_of_size, 
            f'{fileprefix}_{size}.pairs',
            size
        )
    
    write_bed(
        sized_clusters, 
        f'{fileprefix}.bed',
        min_c_size,
        max_c_size
    )

    return stats


def write_stats(stats, filename):
    with open(filename, 'w') as file:
        for size, num in stats.items():
            file.write(
                f'{size}\t{num}\n'
            )



from .extractbc import extract_barcodes, add_extractbc
from .combine import combine_coolers, add_combine
from .annotate import annotate_coolers, add_annotate
from .pairs import make_pairs, add_pairs
from .balance import balance_mcool, add_balance

import logging

import argparse as ap


logging.basicConfig(
    format='%(asctime)s - %(message)s',
    level=logging.INFO
)


cmds = {
    'combine': combine_coolers,
    'annotate': annotate_coolers,
    'extractbc': extract_barcodes,
    'pairs': make_pairs,
    'balance': balance_mcool
}


def parse_args():
    parser = ap.ArgumentParser()
    sub = parser.add_subparsers(dest = 'subcommand')
    add_combine(sub)
    add_annotate(sub)
    add_extractbc(sub)
    add_pairs(sub)
    add_balance(sub)
    return parser.parse_args()


def main():
    args = parse_args()
    cmds[args.subcommand](args)


if __name__ == '__main__':
    main()



def add_extractbc(subparsers):
    parser = subparsers.add_parser(
        'extractbc',
        help = '''
        extract barcode sequences from raw SPRITE-seq reads and append them to the readname.
        by default this only writes reads having complete barcode sequences as defined by the given layouts
        '''
    )
    parser.add_argument(
        '--read1',
        '-r1',
        help = '(gzipped) fastq file containing sequence data for read1',
        required = True
    )
    parser.add_argument(
        '--read2',
        '-r2',
        help = '(gzipped) fastq file containing sequence data for read2',
        required = True
    )
    parser.add_argument(
        '--barcodes',
        '-bc',
        help = 'tab-separated file containing barcode information with columns category, bcname, bcseq',
        required = True
    )
    parser.add_argument(
        '--layout1',
        '-l1',
        help = 'barcode layout for read1 of the form category1|category2|...',
        default = ''
    )
    parser.add_argument(
        '--layout2',
        '-l2',
        help = 'barcode layout for read2 of the form category1|category2|...',
        default = ''
    )
    parser.add_argument(
        '--spacerlen',
        help = 'length of the spacer sequences if used',
        type = int,
        default = 6
    )
    parser.add_argument(
        '--laxity',
        help = 'number of bases to read into the current part of the read for matching barcodes',
        type = int,
        default = 6
    )
    parser.add_argument(
        '--mismatches',
        '-m',
        help = 'number of allowed mismatches per barcode category of the form category1:m1,category2:m2,...',
        required = True
    )
    parser.add_argument(
        '--output',
        '-o',
        help = 'file to write processed reads to',
        required = True
    )
    parser.add_argument(
        '--writefiltered',
        help = 'if set, writes reads with incomplete barcode set to a separate file',
        default = False,
        action = 'store_true'
    )
    parser.add_argument(
        '--writer2',
        help = 'if set, also writes r2 file which is usually only needed for barcode extraction',
        default = False,
        action = 'store_true'
    )
    parser.add_argument(
        '--processes',
        '-p',
        help = '''
        the number of processes to use for processing the reads. amounts to p - 2 extraction threads. e.g. 
        if p = 4 we have one main thread 2 extraction threads and 1 writer thread. if p = 1 no additional threads are spawned
        ''',
        default = 1,
        type = int
    )
    


import gc
import cooler
import h5py

import numpy as np

from scipy.sparse import csr_matrix, triu, tril
from cooler.util import parse_cooler_uri


def attrs_to_dict(h5file):
    attrs = {
        key: h5file.attrs[key]
        for key
        in h5file.attrs.keys()
    }
    return attrs


def copy_attrs(source, dest):
    with (
        h5py.File(source, 'r+') as infile,
        h5py.File(dest, 'r+') as ofile
    ):
        attrs = attrs_to_dict(infile)
        ofile.attrs.update(attrs)


def check_weight(cooleruri, weight_name):
    '''
    checks if weight_name already exist in cooler file

    :param cooleruri:   uri to a given cooleruri
    :param weight_name: name of the weight to check for

    :return:            True if weight already in cooler else False
    '''

    cool_path, group_path = parse_cooler_uri(cooleruri)
    weight_exists = False
    with h5py.File(cool_path, 'r+') as h5:
        grp = h5[group_path]
        if grp['bins'].get(weight_name):
            weight_exists = True

    return weight_exists


def store_weights(cooleruri, bias, weightname, stats = False, overwrite = False):
    '''
    stores an iterable of values as a new weight column in the given cooleruri
    with name set to wightname. code taken from cooler's cooler balance see also
    https://github.com/mirnylab/cooler/blob/master/cooler/cli/balance.py

    :param cooleruri:   uri to a given cooler
    :param bias:        iterable containing balancing weights for each genomic bin
    :param weightname:  name of the weight column

    :return:            None
    '''
    cool_path, group_path = parse_cooler_uri(cooleruri)
    with h5py.File(cool_path, 'r+') as h5:
        grp = h5[group_path]
        # add the bias column to the file
        h5opts = dict(compression='gzip', compression_opts=6)
        if overwrite and weightname in grp['bins']:
            del grp['bins'][weightname]
            
        grp['bins'].create_dataset(weightname, data=bias, **h5opts)

        if stats:
            grp["bins"][weightname].attrs.update(stats)
            


def get_resolutons(coolerpath):
    '''
    returns all resolutions present in a MultiCooler file
    :param coolerpath:  path to MultiCooler file

    :return:            list of strings denoting the resolutions present
    '''
    with h5py.File(coolerpath, 'r+') as h5:
        return list(h5['resolutions'].keys())


def rename_weights(cooleruri, name_map):
    cool_path, group_path = parse_cooler_uri(cooleruri)
    with h5py.File(cool_path, 'r+') as h5:
        grp = h5[group_path]
        h5opts = dict(compression='gzip', compression_opts=6)
        for old_name, new_name in name_map.items():
            # add the bias column to the file
            weights = grp['bins'][old_name][()].copy()
            grp['bins'].create_dataset(new_name, data=weights, **h5opts)
            del grp['bins'][old_name]


def cooler_to_csr(cooleruri):
    '''
    loads a cooler into a csr matrix
    taken from HiCMatrix cool.py see also
    https://github.com/deeptools/HiCMatrix/blob/master/hicmatrix/lib/cool.py

    :param cooleruri:   uri to a given cooler

    :return:            data in cooler as scipy.sparse.csr_matrix
    '''
    cooler_file = cooler.Cooler(cooleruri)
    matrixDataFrame = cooler_file.matrix(
        balance=False, 
        sparse=True, 
        as_pixels=True
    )
    used_dtype = np.int32

    if np.iinfo(np.int32).max < cooler_file.info['nbins']:
        used_dtype = np.int64

    count_dtype = matrixDataFrame[0]['count'].dtype
    data = np.empty(
        cooler_file.info['nnz'], 
        dtype=count_dtype
    )
    instances = np.empty(
        cooler_file.info['nnz'], 
        dtype=used_dtype
    )
    features = np.empty(
        cooler_file.info['nnz'], 
        dtype=used_dtype
    )

    i = 0
    size = cooler_file.info['nbins'] // 32
    if size == 0:
        size = 1

    start_pos = 0
    while i < cooler_file.info['nbins']:
        matrixDataFrameChunk = matrixDataFrame[i:i + size]
        _data = matrixDataFrameChunk['count'].values.astype(count_dtype)
        _instances = matrixDataFrameChunk['bin1_id'].values.astype(used_dtype)
        _features = matrixDataFrameChunk['bin2_id'].values.astype(used_dtype)

        data[start_pos:start_pos + len(_data)] = _data
        instances[start_pos:start_pos + len(_instances)] = _instances
        features[start_pos:start_pos + len(_features)] = _features

        start_pos += len(_features)
        i += size

        del _data
        del _instances
        del _features

    matrix = csr_matrix(
        (data, (instances, features)),
        shape = (
            used_dtype(cooler_file.info['nbins']), 
            used_dtype(cooler_file.info['nbins'])
        ), 
        dtype = count_dtype
    )

    del data
    del instances
    del features
    gc.collect()

    # filling lower triangle in case only upper triangle was saved
    if tril(matrix, k=-1).sum() == 0:
        # this case means that the lower triangle of the
        # symmetric matrix (below the main diagonal)
        # is zero. In this case, replace the lower
        # triangle using the upper triangle
        matrix = matrix + triu(matrix, 1).T

    return matrix



import logging

import pandas as pd

from cooler import Cooler, fileops
from .core import annotate_bins


def annotate_cool(coolpath, bedpaths, outprefix):
    cooler = Cooler(coolpath)

    annotated_bins = cooler.bins()[:].loc[:, ['chrom', 'start', 'end']]
    for bedpath in bedpaths:
        logging.info(f'annotating bins of {coolpath} with clusters from {bedpath}')
        tmp = annotate_bins(cooler, bedpath)

        annotated_bins = annotated_bins.merge(
            tmp,
            on = ['chrom', 'start', 'end'],
            how = 'left'
        )

    outfile = outprefix + '.tsv.gz'
    logging.info(f'writing annotated data to {outfile}')
    annotated_bins.to_csv(
        outfile,
        sep = '\t',
        index = False,
        compression = 'gzip'
    )


def annotate_mcool(mcoolpath, bedpaths, outprefix):
    for coolpath in fileops.list_coolers(mcoolpath):
        uri = mcoolpath + '::' + coolpath
        annotate_cool(
            uri,
            bedpaths,
            outprefix + coolpath.replace('/', '_')
        )


def main(args):
    if fileops.is_multires_file(args.input):
        logging.info('annotating multires cooler')
        annotate_mcool(args.input, args.bed, args.outprefix)

    else:
        logging.info('annotating single cooler')
        annotate_cool(args.input, args.bed, args.outprefix)
